#include "objects_detection_lib.hpp"

#include "video_input/preprocessing/AddBorderFunctor.hpp"

#include "objects_detection/AbstractObjectsDetector.hpp"
#include "objects_detection/ObjectsDetectorFactory.hpp"

#include "video_input/AbstractVideoInput.hpp"
#include "video_input/MetricCamera.hpp"

#include "stereo_matching/ground_plane/AbstractGroundPlaneEstimator.hpp"
#include "stereo_matching/stixels/AbstractStixelWorldEstimator.hpp"

#include "ObjectsDetectionLibGui.hpp"

#include "helpers/get_option_value.hpp"
#include "helpers/data/DataSequence.hpp"
#include "helpers/Log.hpp"
#include "helpers/replace_environment_variables.hpp"

#include "objects_detection/detections.pb.h"

#include <boost/scoped_ptr.hpp>
#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/format.hpp>
#include <boost/thread.hpp>

#include <boost/gil/extension/io/png_io.hpp>

#include <iostream>
#include <fstream>


namespace objects_detection {

using namespace std;
using namespace boost;
using namespace doppia;


scoped_ptr<AddBorderFunctor> add_border_p;

shared_ptr<AbstractObjectsDetector> objects_detector_p;

scoped_ptr<ObjectsDetectionLibGui> gui_p;

typedef DataSequence<doppia_protobuf::Detections> DetectionsDataSequence;
scoped_ptr<DetectionsDataSequence> detections_data_sequence_p;

stixels_t stixels_from_previous_frame;


/// options are stored for the delayed StixelWorldEstimator instanciation
boost::program_options::variables_map options;

boost::gil::rgb8_image_t::point_t input_dimensions;


shared_ptr<CameraCalibration> camera_calibration_p;
shared_ptr<MetricCamera> metric_camera_p;


bool
first_frame = true;


void get_options_description(boost::program_options::options_description &desc)
{
    desc.add_options()
            ("video_input.additional_border",
             boost::program_options::value<int>()->default_value(0),
             "when using process_folder, will add border to the image to enable detection of cropped pedestrians. "
             "Value is in pixels (e.g. 50 pixels)")


            // we allow input file options, even if we do not use them
            ("video_input.images_folder,i", program_options::value<string>(),
             "path to a directory with monocular images. This option will overwrite left/right_filename_mask values")

            ;

    // Objects detection options --
    desc.add(ObjectsDetectorFactory::get_args_options());

    desc.add(AbstractVideoInput::get_args_options());

    return;
}


boost::program_options::variables_map parse_configuration_file(const boost::filesystem::path &configuration_filepath)
{

    boost::program_options::variables_map options;

    boost::program_options::options_description desc("Allowed options");

    get_options_description(desc);

    const string configuration_filename = configuration_filepath.string();

    if (configuration_filename.empty() == false)
    {
        boost::filesystem::path configuration_file_path(configuration_filename);
        if(boost::filesystem::exists(configuration_file_path) == false)
        {
            cout << "\033[1;31mCould not find the configuration file:\033[0m "
                 << configuration_file_path << endl;
            throw std::runtime_error("objects_detection cannot run without a configuration file");
        }

        try
        {
            fstream configuration_file;
            configuration_file.open(configuration_filename.c_str(), fstream::in);
            boost::program_options::store(
                        boost::program_options::parse_config_file(configuration_file, desc), options);
            configuration_file.close();
        }
        catch (...)
        {
            cout << "\033[1;31mError parsing the configuration file named:\033[0m "
                 << configuration_filename << endl;
            cout << desc << endl;
            throw;
        }

        cout << "Parsed the configuration file " << configuration_filename << std::endl;
    }
    else
    {
        std::runtime_error("Configuration file path is emtpy");
    }


    return options;
}


void init_objects_detection(const boost::filesystem::path configuration_filepath)
{
    const boost::program_options::variables_map options = parse_configuration_file(configuration_filepath);

    init_objects_detection(options);
    return;
}


void init_objects_detection(const boost::program_options::variables_map input_options)
{

    options = input_options; // we store the options for future use

    // setup the logging
    {
        logging::get_log().clear(); // we reset previously existing options

        // set our own stdout rules and set cout as console stream --
        logging::LogRuleSet rules_for_stdout;
        rules_for_stdout.add_rule(logging::ErrorMessage, "*"); // we only print errors

        rules_for_stdout.add_rule(logging::WarningMessage, "*"); // also print warnings

        logging::get_log().set_console_stream(std::cout, rules_for_stdout);

        //logging::log(logging::ErrorMessage, "objects_detection") << "Test error message" << std::endl;
    }


    const int additional_border = get_option_value<int>(options, "video_input.additional_border");

    if(additional_border > 0)
    {
        add_border_p.reset(new AddBorderFunctor(additional_border));
    }


    objects_detector_p.reset(ObjectsDetectorFactory::new_instance(options));

    return;
}


void set_monocular_image(input_image_const_view_t &input_view)
{
    if(first_frame and input_view.size() == 0)
    {
        throw std::runtime_error("objects_detection::set_monocular_image received an image of size 0,0");
    }

    if(add_border_p)
    {
        // add the borders (for Markus and Angelos)
        input_image_const_view_t the_input_view = (*add_border_p)(input_view);
        input_dimensions = the_input_view.dimensions();
        objects_detector_p->set_image(the_input_view);
    }
    else
    {
        input_dimensions = input_view.dimensions();
        objects_detector_p->set_image(input_view);
    }

    if(first_frame)
    {
#if defined(OBJECTS_DETECTION_WITH_UI_LIB)
        boost::shared_ptr<doppia::MetricStereoCamera> stereo_camera_p; // left empty for monocular case
        gui_p.reset(new ObjectsDetectionLibGui(input_view.width(), input_view.height(), stereo_camera_p,
                                               objects_detector_p));
#endif
    }

    if(gui_p)
    {
        gui_p->set_monocular_input(input_view);
    }

    // first frame is set to false in compute()
    return;
}



/// blocking call to compute the detections
void compute()
{
    //const double start_objects_detector_compute_wall_time = omp_get_wtime();
    objects_detector_p->compute();
    //cumulated_objects_detector_compute_time += omp_get_wtime() - start_objects_detector_compute_wall_time;

    if(gui_p)
    {
        gui_p->update();
    }

    first_frame = false;
    return;
}


void shift_detections(detections_t &detections)
{
    const int additional_border = add_border_p->additional_border;

    BOOST_FOREACH(detection_t &detection, detections)
    {
        detection_t::rectangle_t &bb = detection.bounding_box;
        bb.max_corner().x(bb.max_corner().x() - additional_border);
        bb.max_corner().y(bb.max_corner().y() - additional_border);
        bb.min_corner().x(bb.min_corner().x() - additional_border);
        bb.min_corner().y(bb.min_corner().y() - additional_border);
    }
    return;
}


/// returns a copy of the current detections, should not be called if detections_are_ready returns false
const detections_t get_detections()
{
    detections_t the_detections = objects_detector_p->get_detections();

    if(add_border_p)
    {
        shift_detections(the_detections);
        return the_detections;
    }
    else
    {
        return the_detections;
    }
}


void record_detections(const boost::filesystem::path &image_path,
                       const detections_t &the_detections,
                       const int additional_border)
{

    typedef AbstractObjectsDetector::detections_t detections_t;
    typedef AbstractObjectsDetector::detection_t detection_t;

    if(detections_data_sequence_p == false)
    {

        using namespace boost::posix_time;
        const ptime current_time(second_clock::local_time());

        const string recording_path = boost::str( boost::format("%i_%02i_%02i_%i_detections.data_sequence")
                                                  % current_time.date().year()
                                                  % current_time.date().month().as_number()
                                                  % current_time.date().day()
                                                  % current_time.time_of_day().total_seconds() );

        if(boost::filesystem::exists(recording_path) == true)
        {
            // this should never happen
            printf("detections recording path == %s\n", recording_path.c_str());
            throw std::runtime_error("Detections recording path already exists. Please wait one second and try again.");
        }

        DetectionsDataSequence::attributes_t attributes;
        attributes.insert(std::make_pair("created_by", "ObjectsDetectionApplication"));

        detections_data_sequence_p.reset(new DetectionsDataSequence(recording_path, attributes));

        printf("\nCreated recording file %s\n\n", recording_path.c_str());
    }

    assert(static_cast<bool>(detections_data_sequence_p) == true);

    DetectionsDataSequence::data_type detections_data;

    detections_data.set_image_name(image_path.string());

    BOOST_FOREACH(const detection_t &detection, the_detections)
    {
        doppia_protobuf::Detection *detection_data_p = detections_data.add_detections();

        detection_data_p->set_score(detection.score);

        // we must do the max(0, value) check since the Point2d only accept positives coordinates,
        // saving a negative value would create an integer overflow
        doppia_protobuf::Point2d &max_corner = *(detection_data_p->mutable_bounding_box()->mutable_max_corner());
        max_corner.set_x(std::max(0, detection.bounding_box.max_corner().x() - additional_border));
        max_corner.set_y(std::max(0, detection.bounding_box.max_corner().y() - additional_border));
        doppia_protobuf::Point2d &min_corner = *(detection_data_p->mutable_bounding_box()->mutable_min_corner());
        min_corner.set_x(std::max(0, detection.bounding_box.min_corner().x() - additional_border));
        min_corner.set_y(std::max(0, detection.bounding_box.min_corner().y() - additional_border));

        doppia_protobuf::Detection::ObjectClasses object_class = doppia_protobuf::Detection::Unknown;
        switch(detection.object_class)
        { // Car, Pedestrian, Bike, Motorbike, Bus, Tram, StaticObject, Unknown

        case detection_t::Car:
            object_class = doppia_protobuf::Detection::Car;
            break;

        case detection_t::Pedestrian:
            object_class = doppia_protobuf::Detection::Pedestrian;
            break;

        case detection_t::Bike:
            object_class = doppia_protobuf::Detection::Bike;
            break;

        case detection_t::Motorbike:
            object_class = doppia_protobuf::Detection::Motorbike;
            break;

        case detection_t::Bus:
            object_class = doppia_protobuf::Detection::Bus;
            break;

        case detection_t::Tram:
            object_class = doppia_protobuf::Detection::Tram;
            break;

        case detection_t::StaticObject:
            object_class = doppia_protobuf::Detection::StaticObject;
            break;

        default:
            throw std::invalid_argument(
                        "ObjectsDetectionApplication::record_detections received a detection "
                        "with an object_class with a no known correspondence in "
                        "the protocol buffer format");
            break;
        }

        detection_data_p->set_object_class(object_class);
    } // end of "for each stixel in stixels"


    detections_data_sequence_p->write(detections_data);

    return;
}


} // end of namespace objects_detection
